@@ -18,11 +18,11 @@
 %%   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
 %%   Technologies LLC, and Rabbit Technologies Ltd.
 %%
-%%   Portions created by LShift Ltd are Copyright (C) 2007-2009 LShift
+%%   Portions created by LShift Ltd are Copyright (C) 2007-2010 LShift
 %%   Ltd. Portions created by Cohesive Financial Technologies LLC are
-%%   Copyright (C) 2007-2009 Cohesive Financial Technologies
+%%   Copyright (C) 2007-2010 Cohesive Financial Technologies
 %%   LLC. Portions created by Rabbit Technologies Ltd are Copyright
-%%   (C) 2007-2009 Rabbit Technologies Ltd.
+%%   (C) 2007-2010 Rabbit Technologies Ltd.
 %%
 %%   All Rights Reserved.
 %%
@@ -46,6 +46,7 @@
          build_heartbeat_frame/0]).
 -export([generate_table/1, encode_properties/2]).
 -export([check_empty_content_body_frame_size/0]).
+-export([ensure_content_encoded/1, clear_encoded_content/1]).
 
 -import(lists).
 
@@ -60,9 +61,11 @@
 -spec(build_simple_content_frames/3 ::
       (channel_number(), content(), non_neg_integer()) -> [frame()]).
 -spec(build_heartbeat_frame/0 :: () -> frame()).
--spec(generate_table/1 :: (amqp_table()) -> binary()). 
+-spec(generate_table/1 :: (amqp_table()) -> binary()).
 -spec(encode_properties/2 :: ([amqp_property_type()], [any()]) -> binary()).
 -spec(check_empty_content_body_frame_size/0 :: () -> 'ok').
+-spec(ensure_content_encoded/1 :: (content()) -> encoded_content()).
+-spec(clear_encoded_content/1 :: (content()) -> unencoded_content()).
 
 -endif.
 
@@ -132,67 +135,83 @@ create_frame(TypeInt, ChannelInt, Payload) ->
 %% I, D, T and F, as well as the QPid extensions b, d, f, l, s, t, x,
 %% and V.
 
-table_field_to_binary({FName, longstr, Value}) ->
-    [short_string_to_binary(FName), "S", long_string_to_binary(Value)];
+table_field_to_binary({FName, Type, Value}) ->
+    [short_string_to_binary(FName) | field_value_to_binary(Type, Value)].
 
-table_field_to_binary({FName, signedint, Value}) ->
-    [short_string_to_binary(FName), "I", <<Value:32/signed>>];
+field_value_to_binary(longstr, Value) ->
+    ["S", long_string_to_binary(Value)];
 
-table_field_to_binary({FName, decimal, {Before, After}}) ->
-    [short_string_to_binary(FName), "D", Before, <<After:32>>];
+field_value_to_binary(signedint, Value) ->
+    ["I", <<Value:32/signed>>];
 
-table_field_to_binary({FName, timestamp, Value}) ->
-    [short_string_to_binary(FName), "T", <<Value:64>>];
+field_value_to_binary(decimal, {Before, After}) ->
+    ["D", Before, <<After:32>>];
 
-table_field_to_binary({FName, table, Value}) ->
-    [short_string_to_binary(FName), "F", table_to_binary(Value)];
+field_value_to_binary(timestamp, Value) ->
+    ["T", <<Value:64>>];
 
-table_field_to_binary({FName, byte, Value}) ->
-    [short_string_to_binary(FName), "b", <<Value:8/unsigned>>];
+field_value_to_binary(table, Value) ->
+    ["F", table_to_binary(Value)];
 
-table_field_to_binary({FName, double, Value}) ->
-    [short_string_to_binary(FName), "d", <<Value:64/float>>];
+field_value_to_binary(array, Value) ->
+    ["A", array_to_binary(Value)];
 
-table_field_to_binary({FName, float, Value}) ->
-    [short_string_to_binary(FName), "f", <<Value:32/float>>];
+field_value_to_binary(byte, Value) ->
+    ["b", <<Value:8/unsigned>>];
 
-table_field_to_binary({FName, long, Value}) ->
-    [short_string_to_binary(FName), "l", <<Value:64/signed>>];
+field_value_to_binary(double, Value) ->
+    ["d", <<Value:64/float>>];
 
-table_field_to_binary({FName, short, Value}) ->
-    [short_string_to_binary(FName), "s", <<Value:16/signed>>];
+field_value_to_binary(float, Value) ->
+    ["f", <<Value:32/float>>];
 
-table_field_to_binary({FName, bool, Value}) ->
-    [short_string_to_binary(FName), "t", if Value -> 1; true -> 0 end];
+field_value_to_binary(long, Value) ->
+    ["l", <<Value:64/signed>>];
 
-table_field_to_binary({FName, binary, Value}) ->
-    [short_string_to_binary(FName), "x", long_string_to_binary(Value)];
+field_value_to_binary(short, Value) ->
+    ["s", <<Value:16/signed>>];
 
-table_field_to_binary({FName, void, _Value}) ->
-    [short_string_to_binary(FName), "V"].
+field_value_to_binary(bool, Value) ->
+    ["t", if Value -> 1; true -> 0 end];
+
+field_value_to_binary(binary, Value) ->
+    ["x", long_string_to_binary(Value)];
+
+field_value_to_binary(void, _Value) ->
+    ["V"].
 
 table_to_binary(Table) when is_list(Table) ->
     BinTable = generate_table(Table),
     [<<(size(BinTable)):32>>, BinTable].
 
+array_to_binary(Array) when is_list(Array) ->
+    BinArray = generate_array(Array),
+    [<<(size(BinArray)):32>>, BinArray].
+
 generate_table(Table) when is_list(Table) ->
     list_to_binary(lists:map(fun table_field_to_binary/1, Table)).
 
+generate_array(Array) when is_list(Array) ->
+    list_to_binary(lists:map(
+                     fun ({Type, Value}) -> field_value_to_binary(Type, Value) end,
+                     Array)).
 
-short_string_to_binary(String) when is_binary(String) and (size(String) < 256) ->
-    [<<(size(String)):8>>, String];
+short_string_to_binary(String) when is_binary(String) ->
+    Len = size(String),
+    if Len < 256 -> [<<(size(String)):8>>, String];
+       true      -> exit(content_properties_shortstr_overflow)
+    end;
 short_string_to_binary(String) ->
     StringLength = length(String),
-    true = (StringLength < 256), % assertion
-    [<<StringLength:8>>, String].
-
+    if StringLength < 256 -> [<<StringLength:8>>, String];
+       true               -> exit(content_properties_shortstr_overflow)
+    end.
 
 long_string_to_binary(String) when is_binary(String) ->
     [<<(size(String)):32>>, String];
 long_string_to_binary(String) ->
     [<<(length(String)):32>>, String].
 
-
 encode_properties([], []) ->
     <<0, 0>>;
 encode_properties(TypeList, ValueList) ->
@@ -224,7 +243,10 @@ encode_properties(Bit, [T | TypeList], [Value | ValueList], FirstShortAcc, Flags
     end.
 
 encode_property(shortstr, String) ->
-    Len = size(String), <<Len:8/unsigned, String:Len/binary>>;
+    Len = size(String),
+    if Len < 256 -> <<Len:8/unsigned, String:Len/binary>>;
+       true      -> exit(content_properties_shortstr_overflow)
+    end;
 encode_property(longstr, String) ->
     Len = size(String), <<Len:32/unsigned, String:Len/binary>>;
 encode_property(octet, Int) ->
@@ -238,32 +260,7 @@ encode_property(longlongint, Int) ->
 encode_property(timestamp, Int) ->
     <<Int:64/unsigned>>;
 encode_property(table, Table) ->
-    encode_table(Table).
-
-
-encode_table(Table) ->
-    TableBin = list_to_binary(lists:map(fun encode_table_entry/1, Table)),
-    Len = size(TableBin),
-    <<Len:32/unsigned, TableBin:Len/binary>>.
-
-
-encode_table_entry({Name, longstr, Value}) ->
-    NLen = size(Name),
-    VLen = size(Value),
-    <<NLen:8/unsigned, Name:NLen/binary, "S", VLen:32/unsigned, Value:VLen/binary>>;
-encode_table_entry({Name, signedint, Value}) ->
-    NLen = size(Name),
-    <<NLen:8/unsigned, Name:NLen/binary, "I", Value:32/signed>>;
-encode_table_entry({Name, decimal, {Before, After}}) ->
-    NLen = size(Name),
-    <<NLen:8/unsigned, Name:NLen/binary, "D", Before:8/unsigned, After:32/unsigned>>;
-encode_table_entry({Name, timestamp, Value}) ->
-    NLen = size(Name),
-    <<NLen:8/unsigned, Name:NLen/binary, "T", Value:64/unsigned>>;
-encode_table_entry({Name, table, Value}) ->
-    NLen = size(Name),
-    TableBin = encode_table(Value),
-    <<NLen:8/unsigned, Name:NLen/binary, "F", TableBin/binary>>.
+    table_to_binary(Table).
 
 check_empty_content_body_frame_size() ->
     %% Intended to ensure that EMPTY_CONTENT_BODY_FRAME_SIZE is
@@ -275,3 +272,19 @@ check_empty_content_body_frame_size() ->
             exit({incorrect_empty_content_body_frame_size,
                   ComputedSize, ?EMPTY_CONTENT_BODY_FRAME_SIZE})
     end.
+
+ensure_content_encoded(Content = #content{properties_bin = PropsBin})
+  when PropsBin =/= 'none' ->
+    Content;
+ensure_content_encoded(Content = #content{properties = Props}) ->
+    Content #content{properties_bin = rabbit_framing:encode_properties(Props)}.
+
+clear_encoded_content(Content = #content{properties_bin = none}) ->
+    Content;
+clear_encoded_content(Content = #content{properties = none}) ->
+    %% Only clear when we can rebuild the properties_bin later in
+    %% accordance to the content record definition comment - maximum
+    %% one of properties and properties_bin can be 'none'
+    Content;
+clear_encoded_content(Content = #content{}) ->
+    Content#content{properties_bin = none}.
