@@ -18,11 +18,11 @@
 %%   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
 %%   Technologies LLC, and Rabbit Technologies Ltd.
 %%
-%%   Portions created by LShift Ltd are Copyright (C) 2007-2009 LShift
+%%   Portions created by LShift Ltd are Copyright (C) 2007-2010 LShift
 %%   Ltd. Portions created by Cohesive Financial Technologies LLC are
-%%   Copyright (C) 2007-2009 Cohesive Financial Technologies
+%%   Copyright (C) 2007-2010 Cohesive Financial Technologies
 %%   LLC. Portions created by Rabbit Technologies Ltd are Copyright
-%%   (C) 2007-2009 Rabbit Technologies Ltd.
+%%   (C) 2007-2010 Rabbit Technologies Ltd.
 %%
 %%   All Rights Reserved.
 %%
@@ -35,7 +35,7 @@
 -include("rabbit_framing.hrl").
 
 -export([recover/0, declare/5, lookup/1, lookup_or_die/1,
-         list/1, info/1, info/2, info_all/1, info_all/2,
+         list/1, info_keys/0, info/1, info/2, info_all/1, info_all/2,
          publish/2]).
 -export([add_binding/4, delete_binding/4, list_bindings/1]).
 -export([delete/2]).
@@ -61,13 +61,14 @@
                             'exchange_not_found' |
                             'exchange_and_queue_not_found'}).
 -spec(recover/0 :: () -> 'ok').
--spec(declare/5 :: (exchange_name(), exchange_type(), bool(), bool(),
+-spec(declare/5 :: (exchange_name(), exchange_type(), boolean(), boolean(),
                     amqp_table()) -> exchange()).
 -spec(check_type/1 :: (binary()) -> atom()).
 -spec(assert_type/2 :: (exchange(), atom()) -> 'ok').
 -spec(lookup/1 :: (exchange_name()) -> {'ok', exchange()} | not_found()).
 -spec(lookup_or_die/1 :: (exchange_name()) -> exchange()).
 -spec(list/1 :: (vhost()) -> [exchange()]).
+-spec(info_keys/0 :: () -> [info_key()]).
 -spec(info/1 :: (exchange()) -> [info()]).
 -spec(info/2 :: (exchange(), [info_key()]) -> [info()]).
 -spec(info_all/1 :: (vhost()) -> [[info()]]).
@@ -79,17 +80,17 @@
 -spec(delete_binding/4 ::
       (exchange_name(), queue_name(), routing_key(), amqp_table()) ->
              bind_res() | {'error', 'binding_not_found'}).
--spec(list_bindings/1 :: (vhost()) -> 
+-spec(list_bindings/1 :: (vhost()) ->
              [{exchange_name(), queue_name(), routing_key(), amqp_table()}]).
 -spec(delete_queue_bindings/1 :: (queue_name()) -> 'ok').
 -spec(delete_transient_queue_bindings/1 :: (queue_name()) -> 'ok').
--spec(topic_matches/2 :: (binary(), binary()) -> bool()).
--spec(headers_match/2 :: (amqp_table(), amqp_table()) -> bool()).
--spec(delete/2 :: (exchange_name(), bool()) ->
+-spec(topic_matches/2 :: (binary(), binary()) -> boolean()).
+-spec(headers_match/2 :: (amqp_table(), amqp_table()) -> boolean()).
+-spec(delete/2 :: (exchange_name(), boolean()) ->
              'ok' | not_found() | {'error', 'in_use'}).
--spec(list_queue_bindings/1 :: (queue_name()) -> 
+-spec(list_queue_bindings/1 :: (queue_name()) ->
               [{exchange_name(), routing_key(), amqp_table()}]).
--spec(list_exchange_bindings/1 :: (exchange_name()) -> 
+-spec(list_exchange_bindings/1 :: (exchange_name()) ->
               [{queue_name(), routing_key(), amqp_table()}]).
 
 -endif.
@@ -165,6 +166,8 @@ list(VHostPath) ->
       rabbit_exchange,
       #exchange{name = rabbit_misc:r(VHostPath, exchange), _ = '_'}).
 
+info_keys() -> ?INFO_KEYS.
+
 map(VHostPath, F) ->
     %% TODO: there is scope for optimisation here, e.g. using a
     %% cursor, parallelising the function invocation
@@ -235,9 +238,9 @@ route(X = #exchange{type = topic}, RoutingKey, _Content) ->
 
 route(X = #exchange{type = headers}, _RoutingKey, Content) ->
     Headers = case (Content#content.properties)#'P_basic'.headers of
-		  undefined -> [];
-		  H         -> sort_arguments(H)
-	      end,
+                  undefined -> [];
+                  H         -> sort_arguments(H)
+              end,
     match_bindings(X, fun (#binding{args = Spec}) ->
                               headers_match(Spec, Headers)
                       end);
@@ -321,7 +324,7 @@ delete_queue_bindings(QueueName, FwdDeleteFun) ->
      end || Route <- mnesia:match_object(
                        rabbit_reverse_route,
                        reverse_route(
-                         #route{binding = #binding{queue_name = QueueName, 
+                         #route{binding = #binding{queue_name = QueueName,
                                                    _ = '_'}}),
                        write)],
     [begin
@@ -430,7 +433,7 @@ list_bindings(VHostPath) ->
     [{ExchangeName, QueueName, RoutingKey, Arguments} ||
         #route{binding = #binding{
                  exchange_name = ExchangeName,
-                 key           = RoutingKey, 
+                 key           = RoutingKey,
                  queue_name    = QueueName,
                  args          = Arguments}}
             <- mnesia:dirty_match_object(
@@ -489,14 +492,14 @@ parse_x_match(Other) ->
 %%
 headers_match(Pattern, Data) ->
     MatchKind = case lists:keysearch(<<"x-match">>, 1, Pattern) of
-		    {value, {_, longstr, MK}} -> parse_x_match(MK);
-		    {value, {_, Type, MK}} ->
-			rabbit_log:warning("Invalid x-match field type ~p "
+                    {value, {_, longstr, MK}} -> parse_x_match(MK);
+                    {value, {_, Type, MK}} ->
+                        rabbit_log:warning("Invalid x-match field type ~p "
                                            "(value ~p); expected longstr",
-					   [Type, MK]),
-			default_headers_match_kind();
-		    _ -> default_headers_match_kind()
-		end,
+                                           [Type, MK]),
+                        default_headers_match_kind();
+                    _ -> default_headers_match_kind()
+                end,
     headers_match(Pattern, Data, true, false, MatchKind).
 
 headers_match([], _Data, AllMatch, _AnyMatch, all) ->
@@ -523,8 +526,8 @@ headers_match([{PK, PT, PV} | PRest], [{DK, DT, DV} | DRest],
             %% the corresponding data field. I've interpreted that to
             %% mean a type of "void" for the pattern field.
             PT == void -> {AllMatch, true};
-	    %% Similarly, it's not specified, but I assume that a
-	    %% mismatched type causes a mismatched value.
+            %% Similarly, it's not specified, but I assume that a
+            %% mismatched type causes a mismatched value.
             PT =/= DT  -> {false, AnyMatch};
             PV == DV   -> {AllMatch, true};
             true       -> {false, AnyMatch}
@@ -597,7 +600,7 @@ list_exchange_bindings(ExchangeName) ->
     [{QueueName, RoutingKey, Arguments} ||
         #route{binding = #binding{queue_name = QueueName,
                                   key = RoutingKey,
-                                  args = Arguments}} 
+                                  args = Arguments}}
             <- mnesia:dirty_match_object(rabbit_route, Route)].
 
 % Refactoring is left as an exercise for the reader
@@ -607,5 +610,5 @@ list_queue_bindings(QueueName) ->
     [{ExchangeName, RoutingKey, Arguments} ||
         #route{binding = #binding{exchange_name = ExchangeName,
                                   key = RoutingKey,
-                                  args = Arguments}} 
+                                  args = Arguments}}
             <- mnesia:dirty_match_object(rabbit_route, Route)].
